import sqlalchemy.orm

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig


def read_experiment_run_phases(
    experiment_id: int = 0,
    experiment_run_id: int = 0,
    experiment_run_instance_id: int = 0,
    as_dict: bool = False,
):
    if experiment_run_instance_id > 0:
        level = 1
    elif experiment_run_id > 0:
        level = 2
    elif experiment_id > 0:
        level = 3
    else:
        level = 0

    store_uri = RuntimeConfig().store_uri

    engine = sqlalchemy.create_engine(store_uri)
    session_maker = sqlalchemy.orm.sessionmaker()
    session_maker.configure(bind=engine)
    raw = []
    with session_maker() as session:
        if level == 0:
            entries = (
                session.query(
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunInstance.id
                    == dbm.ExperimentRunPhase.experiment_run_instance_id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
            )
        elif level == 1:
            entries = (
                session.query(
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunInstance.id
                    == dbm.ExperimentRunPhase.experiment_run_instance_id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(
                    dbm.ExperimentRunInstance.id == experiment_run_instance_id
                )
            )
        elif level == 2:
            entries = (
                session.query(
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunInstance.id
                    == dbm.ExperimentRunPhase.experiment_run_instance_id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(dbm.ExperimentRun.id == experiment_run_id)
            )
        elif level == 3:
            entries = (
                session.query(
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunInstance.id
                    == dbm.ExperimentRunPhase.experiment_run_instance_id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(dbm.ExperimentRun.experiment_id == experiment_id)
            )

        for entry in entries.all():
            raw.append(entry)

    if as_dict:
        results = {
            "phase_id": [],
            "phase_uid": [],
            "phase_mode": [],
            "phase_number": [],
            "instance_id": [],
            "run_id": [],
            "experiment_id": [],
        }
        for entry in raw:
            results["phase_id"].append(entry.ExperimentRunPhase.id)
            results["phase_uid"].append(entry.ExperimentRunPhase.uid)
            results["phase_mode"].append(entry.ExperimentRunPhase.mode)
            results["phase_number"].append(entry.ExperimentRunPhase.number)
            results["instance_id"].append(entry.ExperimentRunInstance.id)
            results["run_id"].append(entry.ExperimentRun.id)
            results["experiment_id"].append(entry.ExperimentRun.experiment_id)
    else:
        results = []
        for entry in raw:
            results.append(
                {
                    "phase_id": entry.ExperimentRunPhase.id,
                    "phase_uid": entry.ExperimentRunPhase.uid,
                    "phase_mode": entry.ExperimentRunPhase.mode,
                    "phase_number": entry.ExperimentRunPhase.number,
                    "instance_id": entry.ExperimentRunInstance.id,
                    "run_id": entry.ExperimentRun.id,
                    "experiment_id": entry.ExperimentRun.experiment_id,
                }
            )

    return results



import sqlalchemy.orm

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig


def read_experiment_run_instances(
    experiment_id: int = 0, experiment_run_id: int = 0, as_dict: bool = False
):
    if experiment_run_id > 0:
        level = 1
    elif experiment_id > 0:
        level = 2
    else:
        level = 0

    store_uri = RuntimeConfig().store_uri

    engine = sqlalchemy.create_engine(store_uri)
    session_maker = sqlalchemy.orm.sessionmaker()
    session_maker.configure(bind=engine)
    raw = []
    with session_maker() as session:
        if level == 0:
            entries = (
                session.query(dbm.ExperimentRunInstance, dbm.ExperimentRun)
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .order_by(dbm.ExperimentRunInstance.id.asc())
            )
        elif level == 1:
            entries = (
                session.query(dbm.ExperimentRunInstance, dbm.ExperimentRun)
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(dbm.ExperimentRun.id == experiment_run_id)
                .order_by(dbm.ExperimentRunInstance.id.asc())
            )
        else:
            entries = (
                session.query(dbm.ExperimentRunInstance, dbm.ExperimentRun)
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(dbm.ExperimentRun.experiment_id == experiment_id)
                .order_by(dbm.ExperimentRunInstance.id.asc())
            )
        # if experiment_id > 0:
        #     runs = (
        #         session.query(dbm.ExperimentRunInstance)
        #         .where(dbm.ExperimentRun.experiment_id == experiment_id)
        #         .order_by(dbm.ExperimentRun.id.asc())
        #     )
        # else:
        #     runs = session.query(dbm.ExperimentRun).order_by(
        #         dbm.ExperimentRun.id.asc()
        #     )
        for entry in entries.all():
            raw.append(entry)

    if as_dict:
        results = {
            "instance_id": [],
            "instance_uid": [],
            "run_id": [],
            "experiment_id": [],
        }
        for entry in raw:
            results["instance_id"].append(entry.ExperimentRunInstance.id)
            results["instance_uid"].append(entry.ExperimentRunInstance.uid)
            results["run_id"].append(entry.ExperimentRun.id)
            results["experiment_id"].append(entry.ExperimentRun.experiment_id)
    else:
        results = []
        for entry in raw:
            results.append(
                {
                    "instance_id": entry.ExperimentRunInstance.id,
                    "instance_uid": entry.ExperimentRunInstance.uid,
                    "run_id": entry.ExperimentRun.id,
                    "experiment_id": entry.ExperimentRun.experiment_id,
                }
            )

    return results



import sqlalchemy.orm

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig


def read_experiments(as_dict: bool = False):
    store_uri = RuntimeConfig().store_uri

    engine = sqlalchemy.create_engine(store_uri)
    session_maker = sqlalchemy.orm.sessionmaker()
    session_maker.configure(bind=engine)
    raw = []
    with session_maker() as session:
        experiments = session.query(dbm.Experiment).order_by(
            dbm.Experiment.id.asc()
        )
        for exp in experiments.all():
            raw.append(exp)

    if as_dict:
        results = {"experiment_id": [], "experiment_name": []}
        for entry in raw:
            results["experiment_id"].append(entry.id)
            results["experiment_name"].append(entry.name)
    else:
        results = []
        for entry in raw:
            results.append(
                {"experiment_id": entry.id, "experiment_name": entry.name}
            )

    return results



import json

import click
import numpy as np
import pandas as pd
import sqlalchemy.orm
from tabulate import tabulate

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig


def read_muscle_actions(
    experiment_id: int = 0,
    experiment_run_id: int = 0,
    experiment_run_instance_id: int = 0,
    experiment_run_phase_id: int = 0,
    limit: int = 10000,
    offset: int = 0,
    as_dict: bool = False,
):
    if experiment_run_phase_id > 0:
        level = 1
    elif experiment_run_instance_id > 0:
        level = 2
    elif experiment_run_id > 0:
        level = 3
    elif experiment_id > 0:
        level = 4
    else:
        level = 0

    store_uri = RuntimeConfig().store_uri

    engine = sqlalchemy.create_engine(store_uri)
    session_maker = sqlalchemy.orm.sessionmaker()
    session_maker.configure(bind=engine)
    raw = []
    with session_maker() as session:
        if level == 0:
            entries = (
                session.query(
                    dbm.MuscleAction,
                    dbm.Agent,
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(dbm.Agent, dbm.MuscleAction.agent_id == dbm.Agent.id)
                .join(
                    dbm.ExperimentRunPhase,
                    dbm.Agent.experiment_run_phase_id
                    == dbm.ExperimentRunPhase.id,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == dbm.ExperimentRunInstance.id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .order_by(dbm.MuscleAction.id.asc())
                .limit(limit)
                .offset(offset)
            )
        elif level == 1:
            entries = (
                session.query(
                    dbm.MuscleAction,
                    dbm.Agent,
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(dbm.Agent, dbm.MuscleAction.agent_id == dbm.Agent.id)
                .join(
                    dbm.ExperimentRunPhase,
                    dbm.Agent.experiment_run_phase_id
                    == dbm.ExperimentRunPhase.id,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == dbm.ExperimentRunInstance.id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(
                    dbm.Agent.experiment_run_phase_id
                    == experiment_run_phase_id
                )
                .order_by(dbm.MuscleAction.id.asc())
                .limit(limit)
                .offset(offset)
            )
        elif level == 2:
            entries = (
                session.query(
                    dbm.MuscleAction,
                    dbm.Agent,
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(dbm.Agent, dbm.MuscleAction.agent_id == dbm.Agent.id)
                .join(
                    dbm.ExperimentRunPhase,
                    dbm.Agent.experiment_run_phase_id
                    == dbm.ExperimentRunPhase.id,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == dbm.ExperimentRunInstance.id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == experiment_run_instance_id
                )
                .order_by(dbm.MuscleAction.id.asc())
                .limit(limit)
                .offset(offset)
            )
        elif level == 3:
            entries = (
                session.query(
                    dbm.MuscleAction,
                    dbm.Agent,
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(dbm.Agent, dbm.MuscleAction.agent_id == dbm.Agent.id)
                .join(
                    dbm.ExperimentRunPhase,
                    dbm.Agent.experiment_run_phase_id
                    == dbm.ExperimentRunPhase.id,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == dbm.ExperimentRunInstance.id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(
                    dbm.ExperimentRunInstance.experiment_run_id
                    == experiment_run_id
                )
                .order_by(dbm.MuscleAction.id.asc())
                .limit(limit)
                .offset(offset)
            )
        elif level == 4:
            entries = (
                session.query(
                    dbm.MuscleAction,
                    dbm.Agent,
                    dbm.ExperimentRunPhase,
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRun,
                )
                .join(dbm.Agent, dbm.MuscleAction.agent_id == dbm.Agent.id)
                .join(
                    dbm.ExperimentRunPhase,
                    dbm.Agent.experiment_run_phase_id
                    == dbm.ExperimentRunPhase.id,
                )
                .join(
                    dbm.ExperimentRunInstance,
                    dbm.ExperimentRunPhase.experiment_run_instance_id
                    == dbm.ExperimentRunInstance.id,
                )
                .join(
                    dbm.ExperimentRun,
                    dbm.ExperimentRunInstance.experiment_run_id
                    == dbm.ExperimentRun.id,
                )
                .where(dbm.ExperimentRun.experiment_id == experiment_id)
                .order_by(dbm.MuscleAction.id.asc())
                .limit(limit)
                .offset(offset)
            )

        for entry in entries.all():
            raw.append(entry)

    if as_dict:
        results = _extract(raw)
    else:
        results = []
        for entry in raw:
            results.append(
                {
                    "muscle_action_id": entry.MuscleAction.id,
                    "agent_id": entry.Agent.id,
                    "phase_id": entry.ExperimentRunPhase.id,
                    "instance_id": entry.ExperimentRunInstance.id,
                    "run_id": entry.ExperimentRun.id,
                    "experiment_id": entry.ExperimentRun.experiment_id,
                    "walltime": entry.MuscleAction.walltime.strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                    "objective": entry.MuscleAction.objective,
                }
            )

    return results


def _extract(data):
    data_dict = {
        "muscle_action_id": [],
        "agent_id": [],
        "agent_name": [],
        "phase_id": [],
        "phase_mode": [],
        "instance_id": [],
        "run_id": [],
        "experiment_id": [],
        "walltime": [],
        "objective": [],
    }

    for idx, entry in enumerate(data):
        data_dict["muscle_action_id"].append(entry.MuscleAction.id)
        data_dict["agent_id"].append(entry.Agent.id)
        data_dict["agent_name"].append(entry.Agent.name)
        data_dict["phase_id"].append(entry.ExperimentRunPhase.id)
        data_dict["phase_mode"].append(entry.ExperimentRunPhase.mode)
        data_dict["instance_id"].append(entry.ExperimentRunInstance.id)
        data_dict["run_id"].append(entry.ExperimentRun.id)
        data_dict["experiment_id"].append(entry.ExperimentRun.experiment_id)
        data_dict["walltime"].append(
            entry.MuscleAction.walltime.strftime("%Y-%m-%d %H:%M:%S")
        )
        data_dict["objective"].append(entry.MuscleAction.objective)

        # Extract simulation times
        for key in entry.MuscleAction.simtimes:
            if key in ("py/object", "default_factory"):
                continue

            env_time_key = f"simtime_{key}"
            env_ticks_key = f"simticks_{key}"
            data_dict.setdefault(env_time_key, [])
            data_dict.setdefault(env_ticks_key, [])
            data_dict[env_time_key].append(
                entry.MuscleAction.simtimes[key]["py/state"][
                    "simtime_timestamp"
                ]
            )
            data_dict[env_ticks_key].append(
                entry.MuscleAction.simtimes[key]["py/state"]["simtime_ticks"]
            )

        try:
            for reading in entry.MuscleAction.sensor_readings:
                sensor_id = ("sensor", reading["py/state"]["uid"])
                data_dict.setdefault(sensor_id, [])
                for idx in range(max(0, idx - len(data_dict[sensor_id]))):
                    data_dict[sensor_id].append(np.nan)

                data_dict[sensor_id].append(reading["py/state"]["value"])

        except TypeError:
            pass

        try:
            for setpoint in entry.MuscleAction.actuator_setpoints:
                actuator_id = ("actuator", setpoint["py/state"]["uid"])
                data_dict.setdefault(actuator_id, [])
                for idx in range(max(0, idx - len(data_dict[actuator_id]))):
                    data_dict[actuator_id].append(np.nan)

                data_dict[actuator_id].append(setpoint["py/state"]["value"])

        except TypeError:
            pass

        try:
            for reward in entry.MuscleAction.rewards:
                reward_id = ("reward", reward["py/state"]["uid"])
                data_dict.setdefault(reward_id, [])
                for idx in range(max(0, idx - len(data_dict[reward_id]))):
                    data_dict[reward_id].append(np.nan)

                data_dict[reward_id].append(reward["py/state"]["value"])

        except TypeError:
            pass
        # for k, v in data_dict.items():
        #    if len(v) < len(data_dict["phase"]):
        #        data_dict[k].append(np.nan)

    # df = pd.concat([df, pd.DataFrame(data_dict)], ignore_index=True)

    # df = pd.concat([df, pd.DataFrame(data_dict)], ignore_index=True)
    for key, val in data_dict.items():
        while len(val) < len(data):
            val.append(np.nan)

    return data_dict



import time
import csv


class BatchReader:
    def __init__(
        self,
        max_read: int = 0,
        start_limit: int = 0,
        start_offset: int = 0,
        to_csv=None,
    ):
        self._max_read = max_read
        self._start_limit = start_limit
        self._base_limit: int = (
            start_limit
            if start_limit > 0
            else (100 if to_csv is None else 10000)
        )
        self._limit = self._base_limit
        self._start_offset = start_offset
        self._offset: int = start_offset
        self._total_lines_read: int = 0
        self._to_csv = to_csv
        self._reading_durations = []
        self._writing_durations = []
        self._results = {}
        self._header = []
        self._last_lines_read: int = 0

    def read_next(self, read_fnc, **fnc_kwargs):
        if not self._continue():
            return {}

        start = time.time()
        fnc_kwargs["limit"] = self._limit
        fnc_kwargs["offset"] = self._offset
        self._results = read_fnc(**fnc_kwargs)
        self._reading_durations.append(time.time() - start)

        self._header = list(self._results.keys())
        self._last_lines_read = len(self._results[self._header[0]])
        self._total_lines_read += self._last_lines_read

        return self._results

    def write_to_csv(self):
        start = time.time()
        mode = "w" if (self._offset == self._start_offset) else "a"
        with open(self._to_csv, mode, newline=None) as csv_file:
            writer = csv.writer(
                csv_file,
                delimiter=",",
                quotechar='"',
                quoting=csv.QUOTE_MINIMAL,
            )
            if mode == "w":
                writer.writerow(self._header)
            for idx in range(self._last_lines_read):
                row = [v[idx] for v in self._results.values()]
                writer.writerow(row)
        self._writing_durations.append(time.time() - start)

    def stop(self):
        if (
            self._results
            and len(self._results[self._header[0]]) < self._base_limit
        ):
            return True
        else:
            self._offset += self._limit
            return False

    def _continue(self):
        if self._max_read > 0:
            if self._total_lines_read >= self._max_read:
                return False
            elif self._total_lines_read + self._limit >= self._max_read:
                self._limit = self._max_read - self._total_lines_read

        return True

    def last_reading_duration(self):
        if self._reading_durations:
            return self._reading_durations[-1]
        else:
            return 0

    def last_writing_duration(self):
        if self._writing_durations:
            return self._writing_durations[-1]
        else:
            return 0

    def total_reading_duration(self):
        return sum(self._reading_durations)

    def total_writing_duration(self):
        return sum(self._writing_durations)

    @property
    def results(self):
        return self._results

    @property
    def lines_read(self):
        return self._total_lines_read



import sqlalchemy.orm

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig


def read_experiment_runs(experiment_id: int = 0, as_dict: bool = False):
    store_uri = RuntimeConfig().store_uri

    engine = sqlalchemy.create_engine(store_uri)
    session_maker = sqlalchemy.orm.sessionmaker()
    session_maker.configure(bind=engine)
    raw = []
    with session_maker() as session:
        if experiment_id > 0:
            runs = (
                session.query(dbm.ExperimentRun)
                .where(dbm.ExperimentRun.experiment_id == experiment_id)
                .order_by(dbm.ExperimentRun.id.asc())
            )
        else:
            runs = session.query(dbm.ExperimentRun).order_by(
                dbm.ExperimentRun.id.asc()
            )
        for run in runs.all():
            raw.append(run)

    if as_dict:
        results = {
            "run_id": [],
            "run_uid": [],
            "experiment_id": [],
        }
        for entry in raw:
            results["run_id"].append(entry.id)
            results["run_uid"].append(entry.uid)
            results["experiment_id"].append(entry.experiment_id)
    else:
        results = []
        for entry in raw:
            results.append(
                {
                    "run_id": entry.id,
                    "run_uid": entry.uid,
                    "experiment_id": entry.experiment_id,
                }
            )

    return results



import csv
import json
import time
from importlib.metadata import version

import click
from click_aliases import ClickAliasedGroup
import numpy as np
import pandas as pd
import sqlalchemy.orm
from tabulate import tabulate

import palaestrai.core.runtime_config
import palaestrai.store.database_model as dbm
from palaestrai.core import RuntimeConfig

from .api_experiment import read_experiments
from .api_experiment_run import read_experiment_runs
from .api_experiment_run_instance import read_experiment_run_instances
from .api_experiment_run_phase import read_experiment_run_phases
from .api_muscle_action import read_muscle_actions
from .util import BatchReader


@click.group(invoke_without_command=True, cls=ClickAliasedGroup)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    help="Supply custom runtime configuration file. "
    "(Default search path: %s)"
    % (palaestrai.core.runtime_config._RuntimeConfig.CONFIG_FILE_PATHS),
)
@click.option(
    "--store-uri",
    "-s",
    "store_uri",
    help=(
        "Specify a custom store uri, default is None which means to use the "
        "value provided in your runtime configuration file."
    ),
)
@click.version_option(version("paldaque"))
def main(config=None, store_uri=None):
    if config:
        try:
            with open(config, "r") as fp:
                RuntimeConfig().reset()  # Make sure we start fresh!
                RuntimeConfig().load(fp)
        except OSError as e:
            click.echo(
                "ERROR: Could not load config from %s: %s." % (config, e),
                file=sys.stderr,
            )
            exit(1)
    else:
        try:
            RuntimeConfig().load()
        except FileNotFoundError as e:
            click.echo(
                "Please create a runtime config. %s.\n"
                "Will continue with built-in defaults." % e,
                file=sys.stderr,
            )
    if store_uri is not None:
        RuntimeConfig().load({"store_uri": store_uri})

    click.echo(
        f"Trying to read from database with URI {RuntimeConfig().store_uri}"
    )


@main.command(aliases=["e"])
def experiment():
    results = read_experiments(as_dict=True)

    click.echo(tabulate(results, headers="keys", tablefmt="pipe"))


@main.command(aliases=["r"])
@click.option("--experiment-id", "-e", "experiment_id", type=int, default=0)
def experiment_run(experiment_id):
    results = read_experiment_runs(experiment_id, as_dict=True)

    click.echo(tabulate(results, headers="keys", tablefmt="pipe"))


@main.command(aliases=["i"])
@click.option("--experiment-id", "-e", "experiment_id", type=int, default=0)
@click.option(
    "--experiment-run-id", "-r", "experiment_run_id", type=int, default=0
)
def experiment_run_instance(experiment_id, experiment_run_id):
    results = read_experiment_run_instances(
        experiment_id, experiment_run_id, as_dict=True
    )
    click.echo(tabulate(results, headers="keys", tablefmt="pipe"))


@main.command(aliases=["p"])
@click.option("--experiment-id", "-e", "experiment_id", type=int, default=0)
@click.option(
    "--experiment-run-id", "-r", "experiment_run_id", type=int, default=0
)
@click.option(
    "--experiment-run-instance-id",
    "-i",
    "experiment_run_instance_id",
    type=int,
    default=0,
)
def experiment_run_phase(
    experiment_id, experiment_run_id, experiment_run_instance_id
):
    results = read_experiment_run_phases(
        experiment_id,
        experiment_run_id,
        experiment_run_instance_id,
        as_dict=True,
    )
    click.echo(tabulate(results, headers="keys", tablefmt="pipe"))


@main.command(aliases=["ma"])
@click.option("--experiment-id", "-e", "experiment_id", type=int, default=0)
@click.option(
    "--experiment-run-id", "-r", "experiment_run_id", type=int, default=0
)
@click.option(
    "--experiment-run-instance-id",
    "-i",
    "experiment_run_instance_id",
    type=int,
    default=0,
)
@click.option(
    "--experiment-run-phase-id",
    "-p",
    "experiment_run_phase_id",
    type=int,
    default=0,
)
@click.option(
    "--csv",
    "-c",
    "to_csv",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        resolve_path=True,
    ),
)
@click.option(
    "--full-console-output", "-f", "full_console_output", is_flag=True
)
@click.option("--max-read", "-m", "max_read", type=int, default=0)
@click.option("--batch-size", "-b", "start_limit", type=int, default=0)
@click.option("--offset", "-o", "start_offset", type=int, default=0)
def muscle_action(
    experiment_id,
    experiment_run_id,
    experiment_run_instance_id,
    experiment_run_phase_id,
    to_csv=None,
    full_console_output=False,
    max_read=0,
    start_limit=0,
    start_offset=0,
):
    verbose = True
    more_verbose = False
    reader = BatchReader(max_read, start_limit, start_offset, to_csv)

    while True:
        reader.read_next(
            read_muscle_actions,
            experiment_id=experiment_id,
            experiment_run_id=experiment_run_id,
            experiment_run_instance_id=experiment_run_instance_id,
            experiment_run_phase_id=experiment_run_phase_id,
            as_dict=True,
        )
        if more_verbose:
            click.echo(
                "Reading finished after "
                f"{reader.last_reading_duration():.3f} seconds."
            )
        if to_csv is not None:
            reader.write_to_csv()
        else:
            results = reader.results
            if not full_console_output:
                results = {
                    k: v
                    for k, v in results.items()
                    if k
                    in (
                        "muscle_action_id",
                        "agent_id",
                        "agent_name",
                        "phase_id",
                        "phase_mode",
                        "instance_id",
                        "run_id",
                        "experiment_id",
                        "walltime",
                        "objective",
                    )
                }

            click.echo(tabulate(results, headers="keys", tablefmt="pipe"))
            if (
                max_read > 0 and reader.lines_read >= max_read
            ) or not click.confirm(
                "Press ENTER to get more (if possible)", default=True
            ):
                break

        if reader.stop():
            break

    if verbose:
        click.echo(
            f"Total reading duration: {reader.total_reading_duration():.3f} seconds"
        )
        writing_durations = reader.total_writing_duration()
        if writing_durations > 0:
            click.echo(
                f"Total writing duration: {writing_durations:.3f} seconds"
            )
